package myutil;
import java.util.Deque;
import java.util.LinkedList;
import java.util.Random;

import commandcenter.CommandCenter;
import enumerate.Action;
import simulator.Simulator;
import structs.CharacterData;
import structs.FrameData;
import structs.GameData;

/**
 * MCTSで利用するNode
 *
 * @author Taichi Miyazaki
 */
public class Node {

  /** UCTの実行時間 */
  //public static final int UCT_TIME = 165 * 100000;
	public  long UCT_TIME ;
  /** UCB1の定数Cの値 */
  public static final double UCB_C = 3;

  /** 探索する木の深さ */
  public static final int UCT_TREE_DEPTH = 2;

  /** ノードを生成する閾値 */
  public static final int UCT_CREATE_NODE_THRESHOULD = 10;

  /** シミュレーションを行う時間 */
  public static final int SIMULATION_TIME = 60;

  /** 乱数を利用するときに使う */
   Random rnd;

  /** 親ノード */
   Node parent;

  /** 子ノード */
   Node[] children;

  /** ノードの深さ */
   int depth;

  /** ノードが探索された回数 */
   int games;

  /** UCB1値 */
   double ucb;

  /** 評価値 */
   double score;

  /** 選択できる自分の全Action */
   LinkedList<Action> myActions;

  /** 選択できる相手の全Action */
   LinkedList<Action> oppActions;

  /** シミュレーションするときに利用する */
   Simulator simulator;

  /** 探索時に選んだ自分のAction */
   LinkedList<Action> selectedMyActions;

  /** シミュレーションする前の自分のHP */
   int myOriginalHp;

  /** シミュレーションする前の相手のHP */
   int oppOriginalHp;

   FrameData frameData;
   boolean playerNumber;
   CommandCenter commandCenter;
   GameData gameData;

   boolean isCreateNode;

  Deque<Action> mAction;
  Deque<Action> oppAction;




  public Node(long ucttime,FrameData frameData, Node parent, LinkedList<Action> myActions,
      LinkedList<Action> oppActions, GameData gameData, boolean playerNumber,
      CommandCenter commandCenter, LinkedList<Action> selectedMyActions) {
    this(ucttime,frameData, parent, myActions, oppActions, gameData, playerNumber, commandCenter);

    this.selectedMyActions = selectedMyActions;
  }

  public Node(long ucttime,FrameData frameData, Node parent, LinkedList<Action> myActions,
      LinkedList<Action> oppActions, GameData gameData, boolean playerNumber,
      CommandCenter commandCenter) {
	  this.UCT_TIME=ucttime;
    this.frameData = frameData;
    this.parent = parent;
    this.myActions = myActions;
    this.oppActions = oppActions;
    this.gameData = gameData;
    this.simulator = new Simulator(gameData);
    this.playerNumber = playerNumber;
    this.commandCenter = commandCenter;

    this.selectedMyActions = new LinkedList<Action>();

    this.rnd = new Random();
    this.mAction = new LinkedList<Action>();
    this.oppAction = new LinkedList<Action>();

    CharacterData myCharacter = playerNumber ? frameData.getP1() : frameData.getP2();
    CharacterData oppCharacter = playerNumber ? frameData.getP2() : frameData.getP1();
    myOriginalHp = myCharacter.getHp();
    oppOriginalHp = oppCharacter.getHp();

    if (this.parent != null) {
      this.depth = this.parent.depth + 1;
    } else {
      this.depth = 0;
    }
  }

  /**
   * MCTSを行う
   *
   * @return 最終的なノードの探索回数が多いAction
   */
  public Action mcts() {
    // 時間の限り、UCTを繰り返す
    long start = System.nanoTime();
    for (; System.nanoTime() - start <= UCT_TIME;) {
      uct();
    }

    return getBestVisitAction();
  }

  /**
   * プレイアウト(シミュレーション)を行う
   *
   * @return プレイアウト結果の評価値
   */
  public double playout() {

    mAction.clear();
    oppAction.clear();

    for (int i = 0; i < selectedMyActions.size(); i++) {
      mAction.add(selectedMyActions.get(i));
    }

    for (int i = 0; i < 5 - selectedMyActions.size(); i++) {
      mAction.add(myActions.get(rnd.nextInt(myActions.size())));
    }

    for (int i = 0; i < 5; i++) {
    	Action toac;
    	toac=oppActions.get(rnd.nextInt(oppActions.size()));
      oppAction.add(oppActions.get(rnd.nextInt(oppActions.size())));
    }

    FrameData nFrameData =
        simulator.simulate(frameData, playerNumber, mAction, oppAction, SIMULATION_TIME); // シミュレーションを実行

    return getScore(nFrameData);
  }

  /**
   * UCTを行う <br>
   *
   * @return 評価値
   */
  public double uct() {

    Node selectedNode = null;
    double bestUcb;

    bestUcb = -99999;

    for (Node child : this.children) {
      if (child.games == 0) {
        child.ucb = 9999 + rnd.nextInt(50);
      } else {
        child.ucb = getUcb(child.score / child.games, games, child.games);
      }


      if (bestUcb < child.ucb) {
        selectedNode = child;
        bestUcb = child.ucb;
      }

    }

    double score = 0;
    if (selectedNode.games == 0) {
      score = selectedNode.playout();
    } else {
      if (selectedNode.children == null) {
        if (selectedNode.depth < UCT_TREE_DEPTH) {
          if (UCT_CREATE_NODE_THRESHOULD <= selectedNode.games) {
            selectedNode.createNode();
            selectedNode.isCreateNode = true;
            score = selectedNode.uct();
          } else {
            score = selectedNode.playout();
          }
        } else {
          score = selectedNode.playout();
        }
      } else {
        if (selectedNode.depth < UCT_TREE_DEPTH) {
          score = selectedNode.uct();
        } else {
          selectedNode.playout();
        }
      }

    }

    selectedNode.games++;
    selectedNode.score += score;

    if (depth == 0) {
      games++;
    }

    return score;
  }

  /**
   * ノードを生成する
   */
  public void createNode() {

    this.children = new Node[myActions.size()];

    for (int i = 0; i < children.length; i++) {

      LinkedList<Action> my = new LinkedList<Action>();
      for (Action act : selectedMyActions) {
        my.add(act);
      }

      my.add(myActions.get(i));

      children[i] =
          new Node(this.UCT_TIME,frameData, this, myActions, oppActions, gameData, playerNumber, commandCenter,
              my);
    }
  }
  public void createNode(LinkedList<Action> myacs) {

	    this.children = new Node[myacs.size()];

	    for (int i = 0; i < children.length; i++) {

	      LinkedList<Action> my = new LinkedList<Action>();
	      for (Action act : selectedMyActions) {
	        my.add(act);
	      }

	      my.add(myacs.get(i));

	      children[i] =
	          new Node(this.UCT_TIME,frameData, this, myActions, oppActions, gameData, playerNumber, commandCenter,
	              my);
	    }
	  }



  /**
   * 最多訪問回数のノードのActionを返す
   *
   * @return 最多訪問回数のノードのAction
   */
  public Action getBestVisitAction() {

    int selected = -1;
    double bestGames = -9999;

    for (int i = 0; i < children.length; i++) {



      if (bestGames < children[i].games) {
        bestGames = children[i].games;
        selected = i;
      }
    }


return children[selected].selectedMyActions.getFirst();
   // return this.myActions.get(selected);
  }



  /**
   * 評価値を返す
   *
   * @param fd フレームデータ(これにhpとかの情報が入っている)
   * @return 評価値
   */
  public int getScore(FrameData fd) {
    return playerNumber ? (fd.getP1().hp - myOriginalHp) - (fd.getP2().hp - oppOriginalHp) : (fd
        .getP2().hp - myOriginalHp) - (fd.getP1().hp - oppOriginalHp);
  }

  /**
   * 評価値と全プレイアウト試行回数とそのActionのプレイアウト試行回数からUCB1値を返す
   *
   * @param score 評価値
   * @param n 全プレイアウト試行回数
   * @param ni そのActionのプレイアウト試行回数
   * @return UCB1値
   */
  public double getUcb(double score, int n, int ni) {
    return score + UCB_C * Math.sqrt((2 * Math.log(n)) / ni);
  }

  public void printNode(Node node) {
    System.out.println("全試行回数:" + node.games);
    for (int i = 0; i < node.children.length; i++) {
      System.out.println(i + ",回数:" + node.children[i].games + ",深さ:" + node.children[i].depth
          + ",score:" + node.children[i].score / node.children[i].games + ",ucb:"
          + node.children[i].ucb);
    }
    System.out.println("");
    for (int i = 0; i < node.children.length; i++) {
      if (node.children[i].isCreateNode) {
        printNode(node.children[i]);
      }
    }
  }
}
